==> grolifant-core-2.0.0-sources.jar: org/ysb33r/grolifant/api/core/runnable/AbstractJvmModelExecTask.java <==
/*
 * ============================================================================
 * (C) Copyright Schalk W. Cronje 2016 - 2023
 *
 * This software is licensed under the Apache License 2.0
 * See http://www.apache.org/licenses/LICENSE-2.0 for license details
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 * ============================================================================
 */
package org.ysb33r.grolifant.api.core.runnable;

import org.gradle.api.Action;
import org.gradle.api.DefaultTask;
import org.gradle.api.tasks.Internal;
import org.gradle.api.tasks.Nested;
import org.gradle.api.tasks.TaskAction;
import org.gradle.process.ExecResult;
import org.gradle.workers.WorkerExecutor;
import org.ysb33r.grolifant.api.core.CmdlineArgumentSpec;
import org.ysb33r.grolifant.api.core.ProjectOperations;
import org.ysb33r.grolifant.api.core.executable.ProcessExecutionSpec;
import org.ysb33r.grolifant.api.core.jvm.ExecutionMode;
import org.ysb33r.grolifant.api.core.jvm.JavaForkOptionsWithEnvProvider;
import org.ysb33r.grolifant.api.core.jvm.JvmEntryPoint;
import org.ysb33r.grolifant.api.core.jvm.worker.WorkerAppParameterFactory;
import org.ysb33r.grolifant.api.core.jvm.worker.WorkerPromise;
import org.ysb33r.grolifant.api.errors.UnsupportedConfigurationException;
import org.ysb33r.grolifant.api.remote.worker.SerializableWorkerAppParameters;
import org.ysb33r.grolifant.api.remote.worker.WorkerAppExecutorFactory;

import javax.annotation.Nullable;
import java.util.Map;
import java.util.TreeMap;

import static org.ysb33r.grolifant.api.core.jvm.ExecutionMode.JAVA_EXEC;

/**
 * Configures a task that will run an execution in a forked JVM.
 *
 * @author Schalk W. Cronj̧é
 * @since 2.0
 */
abstract public class AbstractJvmModelExecTask<T extends AbstractJvmExecSpec, P extends SerializableWorkerAppParameters>
        extends DefaultTask
        implements ForkedJvmExecutable<T>, JvmExecutionModel {

    /**
     * Sets the execution mode.
     * <p>
     * This allows for JVM tasks to be executed either on a worker queue inside the JVM using an isolated classpath,
     * ouside the JVM in a separate process, OR using a classic {@code javaexec}.
     *
     * <p>
     * If nothing is set, the default is {@link ExecutionMode#JAVA_EXEC}.
     *
     * @param em Execution mode.
     * @throw UnsupportedConfigurationException is mode is illegal within context.
     */
    @Override
    public void setExecutionMode(ExecutionMode em) {
        if (this.illegalExecutionModes.containsKey(em)) {
            throw new UnsupportedConfigurationException(em.name() + " cannot be used because: " + illegalExecutionModes.get(em));
        }
        this.execMode = em;
    }

    /**
     * Set whether other tasks in this project should be blocked until this task's worker queue completed.
     * <p>
     * Ignored if {@code executionMode == JAVA_EXEC}.
     *
     * @param mode {@code true} is blocking is required.
     */
    @Override
    public void setAwaitMode(Boolean mode) {
        this.awaitMode = mode;
    }

    /**
     * Configures a JVM that will be forked.
     *
     * @param configurator An action to configure the JVM's fork options.
     * @return The configured executable
     */
    @Override
    public T jvm(Action<JavaForkOptionsWithEnvProvider> configurator) {
        return (T) getJvmExecSpec().jvm(configurator);
    }

    /**
     * Configures the arguments.
     *
     * @param configurator An action to configure the arguments.
     * @return The configured executable
     */
    @Override
    public T runnerSpec(Action<CmdlineArgumentSpec> configurator) {
        return ((T) (getJvmExecSpec().runnerSpec(configurator)));
    }

    /**
     * Configures the entrypoint for the JVM executable.
     * <p>
     * This includes the classpath, modularity and main class.
     * </p>
     *
     * @param configurator An action to configure the entrypoint
     * @return The configured executable.
     */
    @Override
    public T entrypoint(Action<JvmEntryPoint> configurator) {
        return ((T) (getJvmExecSpec().entrypoint(configurator)));
    }

    /**
     * Configures the stream redirection and exit code checks.
     * <p>
     * This includes the inputs and output streams as well as the exit value.
     * </p>
     *
     * @param configurator An action to configure the execution.
     * @return The configured executable.
     */
    @Override
    public T process(Action<ProcessExecutionSpec> configurator) {
        return ((T) (getJvmExecSpec().process(configurator)));
    }

    /**
     * Base class for executing tasks on the JVM.
     *
     * @param we Can be null, in which case workers will not be supported
     */
    protected AbstractJvmModelExecTask(@Nullable WorkerExecutor we) {
        this.projectOperations = ProjectOperations.find(getProject());
        this.execMode = JAVA_EXEC;
        this.awaitMode = false;
        this.workerExecutor = we;
        this.illegalExecutionModes = new TreeMap<>();
    }

    /**
     * Create a worker app executor factory.
     *
     * @return Instance of an execution factory.
     */
    protected abstract WorkerAppExecutorFactory<P> createExecutorFactory();

    /**
     * Create a worker app parameter factory.
     *
     * @return Instance of a parameter factory.
     */
    protected abstract WorkerAppParameterFactory<P> createParameterFactory();


    @Nested
    protected T getJvmExecSpec() {
        return this.execSpec;
    }

    protected void setExecSpec(T execSpec) {
        this.execSpec = execSpec;
    }

    /**
     * Prevents a specific execution mode from being set.
     *
     * @param mode   Illegal execution mode
     * @param reason Reason why execution mode cannot be used.
     */
    protected void preventExecutionMode(ExecutionMode mode, String reason) {
        this.illegalExecutionModes.put(mode, reason);
    }

    @Internal
    protected ProjectOperations getProjectOperations() {
        return this.projectOperations;
    }

    @TaskAction
    protected void exec() {
        // tag::execution-in-multi-mode[]
        if (execMode.equals(JAVA_EXEC) || this.workerExecutor == null) {
            ExecResult result = projectOperations.javaexec(x -> getJvmExecSpec().copyTo(x));
            result.assertNormalExitValue();
        } else {
            WorkerPromise promise = getJvmExecSpec().submitToWorkQueue(
                    execMode.workerIsolation(),
                    workerExecutor,
                    createExecutorFactory(),
                    createParameterFactory()
            );
            if (awaitMode.equals(true)) {
                promise.await();
            }
        }
        // end::execution-in-multi-mode[]
    }

    private T execSpec;
    private ExecutionMode execMode;

    private Boolean awaitMode;
    private final ProjectOperations projectOperations;

    private final WorkerExecutor workerExecutor;

    private final Map<ExecutionMode, String> illegalExecutionModes;
}

==> grolifant-core-2.0.1-sources.jar: org/ysb33r/grolifant/api/core/runnable/AbstractJvmModelExecTask.java <==
/*
 * ============================================================================
 * (C) Copyright Schalk W. Cronje 2016 - 2023
 *
 * This software is licensed under the Apache License 2.0
 * See http://www.apache.org/licenses/LICENSE-2.0 for license details
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 * ============================================================================
 */
package org.ysb33r.grolifant.api.core.runnable;

import groovy.lang.Closure;
import groovy.lang.DelegatesTo;
import org.gradle.api.Action;
import org.gradle.api.DefaultTask;
import org.gradle.api.tasks.Internal;
import org.gradle.api.tasks.Nested;
import org.gradle.api.tasks.TaskAction;
import org.gradle.process.ExecResult;
import org.gradle.workers.WorkerExecutor;
import org.ysb33r.grolifant.api.core.CmdlineArgumentSpec;
import org.ysb33r.grolifant.api.core.ProjectOperations;
import org.ysb33r.grolifant.api.core.executable.ProcessExecutionSpec;
import org.ysb33r.grolifant.api.core.jvm.ExecutionMode;
import org.ysb33r.grolifant.api.core.jvm.JavaForkOptionsWithEnvProvider;
import org.ysb33r.grolifant.api.core.jvm.JvmEntryPoint;
import org.ysb33r.grolifant.api.core.jvm.worker.WorkerAppParameterFactory;
import org.ysb33r.grolifant.api.core.jvm.worker.WorkerPromise;
import org.ysb33r.grolifant.api.errors.UnsupportedConfigurationException;
import org.ysb33r.grolifant.api.remote.worker.SerializableWorkerAppParameters;
import org.ysb33r.grolifant.api.remote.worker.WorkerAppExecutorFactory;

import javax.annotation.Nullable;
import java.util.Map;
import java.util.TreeMap;

import static org.ysb33r.grolifant.api.core.jvm.ExecutionMode.JAVA_EXEC;

/**
 * Configures a task that will run an execution in a forked JVM.
 *
 * @author Schalk W. Cronj̧é
 * @since 2.0
 */
abstract public class AbstractJvmModelExecTask<T extends AbstractJvmExecSpec, P extends SerializableWorkerAppParameters>
        extends DefaultTask
        implements ForkedJvmExecutable<T>, JvmExecutionModel {

    /**
     * Sets the execution mode.
     * <p>
     * This allows for JVM tasks to be executed either on a worker queue inside the JVM using an isolated classpath,
     * ouside the JVM in a separate process, OR using a classic {@code javaexec}.
     *
     * <p>
     * If nothing is set, the default is {@link ExecutionMode#JAVA_EXEC}.
     *
     * @param em Execution mode.
     * @throw UnsupportedConfigurationException is mode is illegal within context.
     */
    @Override
    public void setExecutionMode(ExecutionMode em) {
        if (this.illegalExecutionModes.containsKey(em)) {
            throw new UnsupportedConfigurationException(em.name() + " cannot be used because: " + illegalExecutionModes.get(em));
        }
        this.execMode = em;
    }

    /**
     * Set whether other tasks in this project should be blocked until this task's worker queue completed.
     * <p>
     * Ignored if {@code executionMode == JAVA_EXEC}.
     *
     * @param mode {@code true} is blocking is required.
     */
    @Override
    public void setAwaitMode(Boolean mode) {
        this.awaitMode = mode;
    }

    /**
     * Configures a JVM that will be forked.
     *
     * @param configurator An action to configure the JVM's fork options.
     * @return The configured executable
     */
    @Override
    public T jvm(Action<JavaForkOptionsWithEnvProvider> configurator) {
        return (T) getJvmExecSpec().jvm(configurator);
    }

    /**
     * Configures a JVM that will be forked.
     *
     * @param configurator A closure to configure the JVM's fork options.
     * @return The configured executable
     */
    @Override
    public T jvm(@DelegatesTo(JavaForkOptionsWithEnvProvider.class) Closure<?> configurator) {
        return (T) getJvmExecSpec().jvm(configurator);
    }

    /**
     * Configures the arguments.
     *
     * @param configurator An action to configure the arguments.
     * @return The configured executable
     */
    @Override
    public T runnerSpec(Action<CmdlineArgumentSpec> configurator) {
        return ((T) (getJvmExecSpec().runnerSpec(configurator)));
    }

    /**
     * Configures the arguments.
     *
     * @param configurator A closure to configure the arguments.
     * @return The configured executable
     */
    @Override
    public T runnerSpec(@DelegatesTo(CmdlineArgumentSpec.class) Closure<?> configurator) {
        return ((T) (getJvmExecSpec().runnerSpec(configurator)));
    }

    /**
     * Configures the entrypoint for the JVM executable.
     * <p>
     * This includes the classpath, modularity and main class.
     * </p>
     *
     * @param configurator An action to configure the entrypoint
     * @return The configured executable.
     */
    @Override
    public T entrypoint(Action<JvmEntryPoint> configurator) {
        return ((T) (getJvmExecSpec().entrypoint(configurator)));
    }

    /**
     * Configures the entrypoint for the JVM executable
     *
     * @param configurator A closure to configure the entrypoint
     * @return The configured executable.
     */
    @Override
    public T entrypoint(@DelegatesTo(JvmEntryPoint.class) Closure<?> configurator) {
        return ((T) (getJvmExecSpec().entrypoint(configurator)));
    }

    /**
     * Configures the stream redirection and exit code checks.
     * <p>
     * This includes the inputs and output streams as well as the exit value.
     * </p>
     *
     * @param configurator An action to configure the execution.
     * @return The configured executable.
     */
    @Override
    public T process(Action<ProcessExecutionSpec> configurator) {
        return ((T) (getJvmExecSpec().process(configurator)));
    }

    /**
     * Configures the stream redirections and exit checks.
     *
     * @param configurator A closure to configure the execution.
     * @return The configured executable.
     */
    @Override
    public T process(@DelegatesTo(ProcessExecutionSpec.class) Closure<?> configurator) {
        return ((T) (getJvmExecSpec().process(configurator)));
    }

    /**
     * Base class for executing tasks on the JVM.
     *
     * @param we Can be null, in which case workers will not be supported
     */
    protected AbstractJvmModelExecTask(@Nullable WorkerExecutor we) {
        this.projectOperations = ProjectOperations.find(getProject());
        this.execMode = JAVA_EXEC;
        this.awaitMode = false;
        this.workerExecutor = we;
        this.illegalExecutionModes = new TreeMap<>();
    }

    /**
     * Create a worker app executor factory.
     *
     * @return Instance of an execution factory.
     */
    protected abstract WorkerAppExecutorFactory<P> createExecutorFactory();

    /**
     * Create a worker app parameter factory.
     *
     * @return Instance of a parameter factory.
     */
    protected abstract WorkerAppParameterFactory<P> createParameterFactory();


    @Nested
    protected T getJvmExecSpec() {
        return this.execSpec;
    }

    protected void setExecSpec(T execSpec) {
        this.execSpec = execSpec;
    }

    /**
     * Prevents a specific execution mode from being set.
     *
     * @param mode   Illegal execution mode
     * @param reason Reason why execution mode cannot be used.
     */
    protected void preventExecutionMode(ExecutionMode mode, String reason) {
        this.illegalExecutionModes.put(mode, reason);
    }

    @Internal
    protected ProjectOperations getProjectOperations() {
        return this.projectOperations;
    }

    @TaskAction
    protected void exec() {
        // tag::execution-in-multi-mode[]
        if (execMode.equals(JAVA_EXEC) || this.workerExecutor == null) {
            ExecResult result = projectOperations.javaexec(x -> getJvmExecSpec().copyTo(x));
            result.assertNormalExitValue();
        } else {
            WorkerPromise promise = getJvmExecSpec().submitToWorkQueue(
                    execMode.workerIsolation(),
                    workerExecutor,
                    createExecutorFactory(),
                    createParameterFactory()
            );
            if (awaitMode.equals(true)) {
                promise.await();
            }
        }
        // end::execution-in-multi-mode[]
    }

    private T execSpec;
    private ExecutionMode execMode;

    private Boolean awaitMode;
    private final ProjectOperations projectOperations;

    private final WorkerExecutor workerExecutor;

    private final Map<ExecutionMode, String> illegalExecutionModes;
}
